import { createSlice, PayloadAction } from '@reduxjs/toolkit';
import type { RootState } from './store';
import { getGraphStatistics, GraphStatistics } from '../../statistics';

export interface GraphQueryResultFromBackendPayload {
  queryID: string;
  result: {
    payload: GraphQueryResultFromBackend;
    type: string;
  };
}

export type NodeAttributes = { [key: string]: unknown };

export interface GraphQueryResultFromBackend {
  nodes: {
    _id: string;
    label?: string;
    attributes: NodeAttributes;
  }[];

  edges: {
    _id: string;
    label?: string;
    attributes: NodeAttributes;
    from: string;
    to: string;
  }[];

  // TODO: Also include type in node and edge
  // TODO: The backend should send all the different entitytypes and relationtypes in the result
}

export type Node = {
  _id: string;
  label: string;
  attributes: NodeAttributes;
  mldata?: any; // FIXME
  /* type: string[]; */
};
export type Edge = {
  attributes: NodeAttributes;
  from: string;
  to: string;
  _id: string;
  label: string;
  /* type: string; */
};

// Define a type for the slice state
export type GraphQueryResult = {
  metaData?: GraphStatistics;
  nodes: Node[];
  edges: Edge[];
  queryingBackend: boolean;
};

// Define the initial state using that type
export const initialState: GraphQueryResult = {
  metaData: undefined,
  nodes: [],
  edges: [],
  queryingBackend: false,
};

export const graphQueryBackend2graphQuery = (payload: GraphQueryResultFromBackend) => {
  // Assign ids to a hidden attribute _id to better identify them
  payload.nodes.forEach((node, index) => {
    node._id = node._id || (node as any).id;
  });
  payload.edges.forEach((edge, index) => {
    edge._id = edge._id || (edge as any).id;
  });

  // Only keep one node and one edge per id. This is also done in the backend, but we do it here as well to be sure.
  const nodeIDs = new Set(payload.nodes.map((node) => node._id));
  const edgeIDs = new Set(payload.edges.map((edge) => edge._id));
  let nodes = [...nodeIDs].map((nodeID) => payload.nodes.find((node) => node._id === nodeID) as unknown as Node);
  let edges = [...edgeIDs].map((edgeID) => payload.edges.find((edge) => edge._id === edgeID) as unknown as Edge);

  nodes = nodes.map((node) => {
    let _node = { ...node };
    // TODO!: only works for neo4j
    let nodeType: string = node._id.split('/')[0];
    let innerLabels = node?.attributes?.labels as string[];

    if (innerLabels?.length > 0) nodeType = innerLabels[0] as string;

    _node.label = nodeType;
    return _node;
  });

  edges = edges.map((edge) => {
    let _edge = { ...edge };
    // TODO!: only works for neo4j
    let edgeType: string = edge._id.split('/')[0];

    if (!_edge._id.includes('/')) edgeType = edge.attributes.Type as string;

    _edge.label = edgeType;
    return _edge;
  });

  const metaData = getGraphStatistics(payload);

  return { metaData, nodes, edges };
};

export const graphQueryResultSlice = createSlice({
  name: 'graphQueryResult',
  initialState,
  reducers: {
    setNewGraphQueryResult: (state, action: PayloadAction<GraphQueryResultFromBackendPayload>) => {
      const payload = action.payload.result.payload;
      const { metaData, nodes, edges } = graphQueryBackend2graphQuery(payload);

      // Assign new state
      state.metaData = metaData;
      state.nodes = nodes;
      state.edges = edges;
      state.queryingBackend = false;
    },
    resetGraphQueryResults: (state) => {
      // Assign new state
      state.metaData = undefined;
      state.nodes = [];
      state.edges = [];
      state.queryingBackend = false;
    },
    queryingBackend: (state) => {
      state.queryingBackend = true;
    },
  },
});

export const { setNewGraphQueryResult, resetGraphQueryResults, queryingBackend } = graphQueryResultSlice.actions;

// Other code such as selectors can use the imported `RootState` type
export const selectGraphQueryResult = (state: RootState) => state.graphQueryResult;
export const selectGraphQueryResultNodes = (state: RootState) => state.graphQueryResult.nodes;
export const selectGraphQueryResultLinks = (state: RootState) => state.graphQueryResult.edges;
export const selectGraphQueryResultMetaData = (state: RootState) => state.graphQueryResult.metaData;

export default graphQueryResultSlice.reducer;
